@@ -1,7 +1,7 @@
 /*******************************************************************************
    This file is part of SecureSkat.
 
- Copyright (C) 2004, 2005  Heiko Stamer <stamer@gaos.org>
+ Copyright (C) 2004, 2005, 2006  Heiko Stamer <stamer@gaos.org>
 
    SecureSkat is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -15,7 +15,7 @@
 
    You should have received a copy of the GNU General Public License
    along with SecureSkat; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
+   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *******************************************************************************/
 
 #include "SecureSkat_vtmf.hh"
@@ -27,7 +27,7 @@ std::string wstr;
 
 // values for biding
 size_t reiz_wert[] =
-	{ 
+	{
 		0, 18, 20, 22, 23, 24, 27, 30, 33, 35, 36, 40, 44, 45, 46, 48, 50, 54, 
 		55, 59, 60, 63, 66, 70, 72, 77, 80, 81, 84, 88, 90, 96, 99, 100, 108, 
 		110, 117, 120, 121, 126, 130, 132, 135, 140, 143, 144, 150, 153, 154, 
@@ -38,11 +38,11 @@ size_t reiz_wert[] =
 size_t pkt_wert[] =
 	{
 		2, 2, 2, 2, 
-		11, 10, 4, 3, 0, 0, 0, 11, 10, 4, 3, 0, 0, 0,
+		11, 10, 4, 3, 0, 0, 0, 11, 10, 4, 3, 0, 0, 0, 
 		11, 10, 4, 3, 0, 0, 0, 11, 10, 4, 3, 0, 0, 0
 	};
 
-// order of the cards in several games
+// order of the cards in several games (99 is a marker for an invalid card)
 size_t card_order[][5][18] =
 	{
 		// Sc
@@ -81,7 +81,7 @@ size_t card_order[][5][18] =
 				0,  1,  2,  3, 18, 19, 20, 21, 22, 23, 24, 99, 99, 99, 99, 99, 99, 99
 			}
 		},
-		// Gr	
+		// Gr
 		{
 			{
 				0,  1,  2,  3, 11, 12, 13, 14, 15, 16, 17,  4,  5,  6,  7,  8,  9, 10
@@ -374,7 +374,6 @@ int skat_vkarte
 	
 	int type = -1;
 	VTMF_Card c;
-	char *tmp = new char[TMCG_MAX_CARD_CHARS];
 	
 	try
 	{
@@ -382,8 +381,8 @@ int skat_vkarte
 			((pkr_self == 1) && (pkr_who == 2)) || 
 			((pkr_self == 2) && (pkr_who == 0)))
 		{
-			left->getline(tmp, TMCG_MAX_CARD_CHARS);
-			if (!c.import(tmp))
+			*left >> c;
+			if (!left->good())
 				throw -1;
 			if (!s.find(c))
 				throw -1;
@@ -416,8 +415,8 @@ int skat_vkarte
 			((pkr_self == 1) && (pkr_who == 0)) || 
 			((pkr_self == 2) && (pkr_who == 1)))
 		{
-			right->getline(tmp, TMCG_MAX_CARD_CHARS);
-			if (!c.import(tmp))
+			*right >> c;
+			if (!right->good())
 				throw -1;
 			if (!s.find(c))
 				throw -1;
@@ -450,7 +449,6 @@ int skat_vkarte
 	}
 	catch (int return_value)
 	{
-		delete [] tmp;
 #ifndef NDEBUG
 		stop_clock();
 		std::cerr << elapsed_time() << std::flush;
@@ -860,37 +858,39 @@ bool skat_geben
 bool skat_mischen_beweis
 	(
 		size_t pkr_self, SchindelhauerTMCG *tmcg, BarnettSmartVTMF_dlog *vtmf,
-		const TMCG_Stack<VTMF_Card> &d, const TMCG_StackSecret<VTMF_CardSecret> &ss,
-		const TMCG_Stack<VTMF_Card> &d0, const TMCG_Stack<VTMF_Card> &d1, const TMCG_Stack<VTMF_Card> &d2,
+		GrothVSSHE *vsshe, const TMCG_Stack<VTMF_Card> &d,
+		const TMCG_StackSecret<VTMF_CardSecret> &ss,
+		const TMCG_Stack<VTMF_Card> &d0, const TMCG_Stack<VTMF_Card> &d1,
+		const TMCG_Stack<VTMF_Card> &d2,
 		iosecuresocketstream *right, iosecuresocketstream *left
 	)
 {
 	if (pkr_self == 0)
 	{
-		tmcg->TMCG_ProveStackEquality(d, d0, ss, false, vtmf, *left, *left);
-		tmcg->TMCG_ProveStackEquality(d, d0, ss, false, vtmf, *right, *right);
-		if (!tmcg->TMCG_VerifyStackEquality(d0, d1, false, vtmf, *left, *left))
+		tmcg->TMCG_ProveStackEquality_Groth(d, d0, ss, vtmf, vsshe, *left, *left);
+		tmcg->TMCG_ProveStackEquality_Groth(d, d0, ss, vtmf, vsshe, *right, *right);
+		if (!tmcg->TMCG_VerifyStackEquality_Groth(d0, d1, vtmf, vsshe, *left, *left))
 			return false;
-		if (!tmcg->TMCG_VerifyStackEquality(d1, d2, false, vtmf, *right, *right))
+		if (!tmcg->TMCG_VerifyStackEquality_Groth(d1, d2, vtmf, vsshe, *right, *right))
 			return false;
 	}
 	if (pkr_self == 1)
 	{
-		if (!tmcg->TMCG_VerifyStackEquality(d, d0, false, vtmf, *right, *right))
+		if (!tmcg->TMCG_VerifyStackEquality_Groth(d, d0, vtmf, vsshe, *right, *right))
 			return false;
-		tmcg->TMCG_ProveStackEquality(d0, d1, ss, false, vtmf, *right, *right);
-		tmcg->TMCG_ProveStackEquality(d0, d1, ss, false, vtmf, *left, *left);
-		if (!tmcg->TMCG_VerifyStackEquality(d1, d2, false, vtmf, *left, *left))
+		tmcg->TMCG_ProveStackEquality_Groth(d0, d1, ss, vtmf, vsshe, *right, *right);
+		tmcg->TMCG_ProveStackEquality_Groth(d0, d1, ss, vtmf, vsshe, *left, *left);
+		if (!tmcg->TMCG_VerifyStackEquality_Groth(d1, d2, vtmf, vsshe, *left, *left))
 			return false;
 	}
 	if (pkr_self == 2)
 	{
-		if (!tmcg->TMCG_VerifyStackEquality(d, d0, false, vtmf, *left, *left))
+		if (!tmcg->TMCG_VerifyStackEquality_Groth(d, d0, vtmf, vsshe, *left, *left))
 			return false;
-		if (!tmcg->TMCG_VerifyStackEquality(d0, d1, false, vtmf, *right, *right))
+		if (!tmcg->TMCG_VerifyStackEquality_Groth(d0, d1, vtmf, vsshe, *right, *right))
 			return false;
-		tmcg->TMCG_ProveStackEquality(d1, d2, ss, false, vtmf, *left, *left);
-		tmcg->TMCG_ProveStackEquality(d1, d2, ss, false, vtmf, *right, *right);
+		tmcg->TMCG_ProveStackEquality_Groth(d1, d2, ss, vtmf, vsshe, *left, *left);
+		tmcg->TMCG_ProveStackEquality_Groth(d1, d2, ss, vtmf, vsshe, *right, *right);
 	}
 	return true;
 }
@@ -904,53 +904,46 @@ bool skat_mischen
 		iosecuresocketstream *right, iosecuresocketstream *left
 	)
 {
-	char *tmp = new char[TMCG_MAX_STACK_CHARS];
+	if (pkr_self == 0)
+	{
+		tmcg->TMCG_MixStack(d, d0, ss, vtmf);
+		*right << d0 << std::endl << std::flush;
+		*left << d0 << std::endl << std::flush;
+		*left >> d1;
+		if (!left->good())
+			return false;
+		*right >> d2;
+		if (!right->good())
+			return false;
+	}
+	else if (pkr_self == 1)
+	{
+		*right >> d0;
+		if (!right->good())
+			return false;
+		tmcg->TMCG_MixStack(d0, d1, ss, vtmf);
+		*right << d1 << std::endl << std::flush;
+		*left << d1 << std::endl << std::flush;
+		*left >> d2;
+		if (!left->good())
+			return false;
+	}
+	else if (pkr_self == 2)
+	{
+		*left >> d0;
+		if (!left->good())
+			return false;
+		*right >> d1;
+		if (!right->good())
+			return false;
+		tmcg->TMCG_MixStack(d1, d2, ss, vtmf);
+		*right << d2 << std::endl << std::flush;
+		*left << d2 << std::endl << std::flush;
+	}
+	else
+		return false;
 	
-	try
-	{
-		if (pkr_self == 0)
-		{
-			tmcg->TMCG_MixStack(d, d0, ss, vtmf);
-			*right << d0 << std::endl << std::flush;
-			*left << d0 << std::endl << std::flush;
-			left->getline(tmp, TMCG_MAX_STACK_CHARS);
-			if (!d1.import(tmp))
-				throw false;
-			right->getline(tmp, TMCG_MAX_STACK_CHARS);
-			if (!d2.import(tmp))
-				throw false;
-		}
-		if (pkr_self == 1)
-		{
-			right->getline(tmp, TMCG_MAX_STACK_CHARS);
-			if (!d0.import(tmp))
-				throw false;
-			tmcg->TMCG_MixStack(d0, d1, ss, vtmf);
-			*right << d1 << std::endl << std::flush;
-			*left << d1 << std::endl << std::flush;
-			left->getline(tmp, TMCG_MAX_STACK_CHARS);
-			if (!d2.import(tmp))
-				throw false;
-		}
-		if (pkr_self == 2)
-		{
-			left->getline(tmp, TMCG_MAX_STACK_CHARS);
-			if (!d0.import(tmp))
-				throw false;
-			right->getline(tmp, TMCG_MAX_STACK_CHARS);
-			if (!d1.import(tmp))
-				throw false;
-			tmcg->TMCG_MixStack(d1, d2, ss, vtmf);
-			*right << d2 << std::endl << std::flush;
-			*left << d2 << std::endl << std::flush;
-		}
-		throw true;
-	}
-	catch (bool return_value)
-	{
-		delete [] tmp;
-		return return_value;
-	}
+	return true;
 }
 
 bool game_helper_1
@@ -973,7 +966,7 @@ bool game_helper_1
 			return false;
 		case 12:
 			reiz_status += 100;
-			std::cout << "><><>< VH aka \"" << pkr.key[vh].name << 
+			std::cout << "><><>< VH aka \"" << pkr.keys[vh].name << 
 				"\" " << _("gets the game at") << " " << 
 				reiz_wert[reiz_counter] << std::endl;
 			spiel_allein = vh;
@@ -983,7 +976,7 @@ bool game_helper_1
 			return true;
 		case 13:
 			reiz_status += 100;
-			std::cout << "><><>< MH aka \"" << pkr.key[mh].name << 
+			std::cout << "><><>< MH aka \"" << pkr.keys[mh].name << 
 				"\" " << _("gets the game at") << " " << 
 				reiz_wert[reiz_counter] << std::endl;
 			spiel_allein = mh;
@@ -993,7 +986,7 @@ bool game_helper_1
 			return true;
 		case 14:
 			reiz_status += 100;
-			std::cout << "><><>< HH aka \"" << pkr.key[hh].name << 
+			std::cout << "><><>< HH aka \"" << pkr.keys[hh].name << 
 				"\" " << _("gets the game at") << " " << 
 				reiz_wert[reiz_counter] << std::endl;
 			spiel_allein = hh;
@@ -1025,12 +1018,13 @@ int skat_game
 	{
 		std::ostringstream ost;
 		ost << nicks[pkr_self] << " INIT " << nicks[i] << " " << 
-			pkr.key[i].name << std::endl;
+			pkr.keys[i].name << std::endl;
 		*out_ctl << ost.str() << std::flush;
 	}
 	
 	// VTMF initalization
 	BarnettSmartVTMF_dlog *vtmf;
+	
 	if (pkr_self == 2)
 	{
 #ifndef NDEBUG
@@ -1039,8 +1033,11 @@ int skat_game
 		
 #ifdef COMMON_DDH_GROUP
 		std::stringstream ddh_group;
-		ddh_group << COMMON_DDH_GROUP << std::endl;
-		vtmf = new BarnettSmartVTMF_dlog(ddh_group);
+		ddh_group << COMMON_DDH_GROUP_P << std::endl;
+		ddh_group << COMMON_DDH_GROUP_Q << std::endl;
+		ddh_group << COMMON_DDH_GROUP_G << std::endl;
+		ddh_group << "2" << std::endl;
+		vtmf = new BarnettSmartVTMF_dlog_GroupQR(ddh_group);
 #else
 		vtmf = new BarnettSmartVTMF_dlog();
 		vtmf->PublishGroup(*left), vtmf->PublishGroup(*right);
@@ -1086,8 +1083,11 @@ int skat_game
 		
 #ifdef COMMON_DDH_GROUP
 		std::stringstream ddh_group;
-		ddh_group << COMMON_DDH_GROUP << std::endl;
-		vtmf = new BarnettSmartVTMF_dlog(ddh_group);
+		ddh_group << COMMON_DDH_GROUP_P << std::endl;
+		ddh_group << COMMON_DDH_GROUP_Q << std::endl;
+		ddh_group << COMMON_DDH_GROUP_G << std::endl;
+		ddh_group << "2" << std::endl;
+		vtmf = new BarnettSmartVTMF_dlog_GroupQR(ddh_group);
 #else
 		if (pkr_self == 0)
 			vtmf = new BarnettSmartVTMF_dlog(*right);
@@ -1147,6 +1147,92 @@ int skat_game
 #endif
 	}
 	
+	// initalization for Groth's shuffle
+	GrothVSSHE *vsshe;
+	
+	if (pkr_self == 2)
+	{
+#ifndef NDEBUG
+		start_clock();
+#endif
+		
+		vsshe = new GrothVSSHE(32, vtmf->p, vtmf->q, vtmf->k, 
+			vtmf->g, vtmf->h);
+		vsshe->PublishGroup(*left), vsshe->PublishGroup(*right);
+		if (!vsshe->CheckGroup())
+		{
+			std::cout << ">< " << _("VSSHE ERROR") << ": " <<
+				_("function CheckGroup() failed") << std::endl;
+			return 2;
+		}
+		if (mpz_cmp(vtmf->h, vsshe->com->h))
+		{
+			std::cout << ">< " << _("VSSHE ERROR") << ": " <<
+				_("common public key does not match") << std::endl;
+			return 2;
+		}
+		if (mpz_cmp(vtmf->q, vsshe->com->q))
+		{
+			std::cout << ">< " << _("VSSHE ERROR") << ": " <<
+				_("subgroup order does not match") << std::endl;
+			return 2;
+		}
+		if (mpz_cmp(vtmf->p, vsshe->p) || mpz_cmp(vtmf->q, vsshe->q) || 
+			mpz_cmp(vtmf->g, vsshe->g) || mpz_cmp(vtmf->h, vsshe->h))
+		{
+			std::cout << ">< " << _("VSSHE ERROR") << ": " <<
+				_("encryption scheme does not match") << std::endl;
+			return 2;
+		}
+		
+#ifndef NDEBUG
+		stop_clock();
+		std::cerr << "KeyGenerationProtocol2: " << elapsed_time() << std::endl;
+#endif
+	}
+	else
+	{
+#ifndef NDEBUG
+		start_clock();
+#endif
+		
+		if (pkr_self == 0)
+			vsshe = new GrothVSSHE(32, *right);
+		else
+			vsshe = new GrothVSSHE(32, *left);
+		if (!vsshe->CheckGroup())
+		{
+			std::cout << ">< " << _("VSSHE ERROR") << ": " <<
+				_("function CheckGroup() failed") << std::endl;
+			return 2;
+		}
+		if (mpz_cmp(vtmf->h, vsshe->com->h))
+		{
+			std::cout << ">< " << _("VSSHE ERROR") << ": " <<
+				_("common public key does not match") << std::endl;
+			return 2;
+		}
+		if (mpz_cmp(vtmf->q, vsshe->com->q))
+		{
+			std::cout << ">< " << _("VSSHE ERROR") << ": " <<
+				_("subgroup order does not match") << std::endl;
+			return 2;
+		}
+		if (mpz_cmp(vtmf->p, vsshe->p) || mpz_cmp(vtmf->q, vsshe->q) || 
+			mpz_cmp(vtmf->g, vsshe->g) || mpz_cmp(vtmf->h, vsshe->h))
+		{
+			std::cout << ">< " << _("VSSHE ERROR") << ": " <<
+				_("encryption scheme does not match") << std::endl;
+			return 2;
+		}
+		
+#ifndef NDEBUG
+		stop_clock();
+		std::cerr << "KeyGenerationProtocol2: " << elapsed_time() << std::endl;
+#endif
+	}
+	
+	// loop the given number of rounds
 	for (size_t r = 0; r < rounds; r++)
 	{
 		std::ostringstream spiel_protokoll;
@@ -1196,7 +1282,7 @@ int skat_game
 #ifndef NDEBUG
 			start_clock();
 #endif
-			if (!skat_mischen_beweis(pkr_self, tmcg, vtmf, d2, ss,
+			if (!skat_mischen_beweis(pkr_self, tmcg, vtmf, vsshe, d2, ss,
 				d_mix[0], d_mix[1], d_mix[2], right, left))
 			{
 				std::cout << ">< " << _("shuffling error") << ": " <<
@@ -1239,8 +1325,7 @@ int skat_game
 			}
 			else
 			{	
-				char *tmp = new char[TMCG_MAX_STACK_CHARS];
-				iosecuresocketstream *hhs;
+				iosecuresocketstream *hhs = NULL;
 				if ((pkr_self == 0) && (p == 0))
 					hhs = right;
 				else if ((pkr_self == 0) && (p == 1))
@@ -1254,13 +1339,13 @@ int skat_game
 				else if ((pkr_self == 2) && (p == 2))
 					hhs = left;
 				else
-					hhs = NULL;
-				hhs->getline(tmp, TMCG_MAX_STACK_CHARS);
-				if (!d_end.import(tmp))
+					return 1;
+				hhs >> d_end;
+				if (!hhs.good())
 				{
 					std::cout << ">< " << _("cutting error") << ": " <<
 						_("bad stack format") << std::endl;
-					delete [] tmp, return 1;
+					return 1;
 				}
 				std::cout << "." << std::flush;
 				if (!tmcg->TMCG_VerifyStackEquality(d_mix[2], d_end,
@@ -1268,9 +1353,8 @@ int skat_game
 				{
 					std::cout << ">< " << _("cutting error") << ": " <<
 						_("wrong ZK proof") << std::endl;
-					delete [] tmp, return 2;
+					return 2;
 				}
-				delete [] tmp;
 			}
 #ifndef NDEBUG
 			stop_clock();
@@ -1520,7 +1604,7 @@ int skat_game
 									break;
 							}
 							std::cout << "><><>< " << _("player") << " \"" << 
-								pkr.key[who_biding].name << "\" " << _("passes at") << " " << 
+								pkr.keys[who_biding].name << "\" " << _("passes at") << " " << 
 								reiz_wert[reiz_counter] << std::endl;
 							if (pctl)
 								*out_ctl << nicks[who_biding] << " PASSE" << std::endl << 
@@ -1595,7 +1679,7 @@ int skat_game
 								
 							}
 							std::cout << "><><>< " << _("player") << " \"" << 
-								pkr.key[who_biding].name << "\" " << _("bids") << " " << 
+								pkr.keys[who_biding].name << "\" " << _("bids") << " " << 
 								reiz_wert[reiz_counter] << std::endl;
 							if (pctl)
 							{
@@ -1617,7 +1701,7 @@ int skat_game
 							{
 								hand_spiel = true, reiz_status += 100;
 								std::cout << "><><>< " << _("player") << " \"" <<
-									pkr.key[spiel_allein].name << 
+									pkr.keys[spiel_allein].name << 
 									"\" " << _("doesn't take the Skat") << std::endl;
 								if (pctl)
 									*out_ctl << nicks[spiel_allein] << " HAND" << std::endl << 
@@ -1633,7 +1717,7 @@ int skat_game
 							{
 								hand_spiel = false, reiz_status += 100;
 								std::cout << "><><>< " << _("player") << " \"" << 
-									pkr.key[spiel_allein].name << 
+									pkr.keys[spiel_allein].name << 
 									"\" " << _("takes the Skat") << std::endl;
 								assert(sk.size() == 2);
 								s[spiel_allein].push(sk);
@@ -1675,12 +1759,10 @@ int skat_game
 								if ((nick == nicks[spiel_allein]) && 
 									(pkr_self != spiel_allein))
 								{
-									char *tmp1 = new char[TMCG_MAX_CARD_CHARS];
-									char *tmp2 = new char[TMCG_MAX_CARD_CHARS];
 									VTMF_Card c1, c2;
 									reiz_status += 100;
 									std::cout << "><><>< " << _("player") << " \"" << 
-										pkr.key[spiel_allein].name << 
+										pkr.keys[spiel_allein].name << 
 										"\" " << _("pushes the Skat") << std::endl;
 									if (pctl)
 										*out_ctl << nicks[spiel_allein] << " DRUECKE" << 
@@ -1689,45 +1771,30 @@ int skat_game
 										((pkr_self == 1) && (spiel_allein == 2)) || 
 										((pkr_self == 2) && (spiel_allein == 0)))
 									{
-										left->getline(tmp1, TMCG_MAX_CARD_CHARS);
-										left->getline(tmp2, TMCG_MAX_CARD_CHARS);
+										*left >> c1 >> c2;
+										if (!left->good())
+											return 9;
 									}
 									else if (((pkr_self == 0) && (spiel_allein == 2)) || 
 										((pkr_self == 1) && (spiel_allein == 0)) || 
 										((pkr_self == 2) && (spiel_allein == 1)))
 									{
-										right->getline(tmp1, TMCG_MAX_CARD_CHARS);
-										right->getline(tmp2, TMCG_MAX_CARD_CHARS);
+										*right >> c1 >> c2;
+										if (!right->good())
+											return 9;
 									}
 									else
-									{
-										delete [] tmp1, delete [] tmp2;
 										return 9;
-									}
 									
 									// check and store pushed cards
-									if (!c1.import(tmp1))
-									{
-										delete [] tmp1, delete [] tmp2;
-										return 9;
-									}
-									if (!c2.import(tmp2))
-									{
-										delete [] tmp1, delete [] tmp2;
-										return 9;
-									}
-									if ((!s[spiel_allein].find(c1)) ||
+									if ((!s[spiel_allein].find(c1)) || 
 										(!s[spiel_allein].find(c2)))
-									{
-										delete [] tmp1, delete [] tmp2;
-										return 10;
-									}
+											return 10;
 									sk.clear();
 									s[spiel_allein].remove(c1);
 									sk.push(c1);
 									s[spiel_allein].remove(c2);
 									sk.push(c2);
-									delete [] tmp1, delete [] tmp2;
 								}
 							}
 						}
@@ -1788,7 +1855,7 @@ int skat_game
 									skat_blatt(10, os_ov);
 								}
 								std::cout << "><><>< " << _("player") << " \"" << 
-									pkr.key[spiel_allein].name << "\" " << _("announces") << 
+									pkr.keys[spiel_allein].name << "\" " << _("announces") << 
 									": " << skat_spiel2string(spiel_status) << std::endl;
 								if (pctl)
 								{
@@ -1818,7 +1885,7 @@ int skat_game
 								return 12;
 							}
 							std::cout << "><><>< " << _("player") << " \"" << 
-								pkr.key[spiel_who[spiel_dran]].name << "\" " << 
+								pkr.keys[spiel_who[spiel_dran]].name << "\" " << 
 								_("plays the card") << ": " <<
 								skat_type2string(type) << std::endl;
 							if (pctl)
@@ -1839,7 +1906,7 @@ int skat_game
 								int bk = skat_bstich(os_sp, spiel_status);
 								assert (bk != -1);
 								std::cout << "><><>< " << _("player") << " \"" << 
-									pkr.key[spiel_who[bk]].name << "\" " << 
+									pkr.keys[spiel_who[bk]].name << "\" " << 
 									_("gets the trick") << ": ";
 								for (size_t i = 0; i < os_sp.size(); i++)
 									std::cout << skat_type2string(os_sp[i].first);
@@ -1861,6 +1928,12 @@ int skat_game
 								spiel_who[1] = (spiel_who[0] + 1) % 3;
 								spiel_who[2] = (spiel_who[0] + 2) % 3;
 								spiel_dran = 0;
+								// Null-Spiele ggf. sofort abbrechen
+								if ((skat_spiel2gwert(spiel_status) == 23)
+									&& (os_pkt[spiel_allein].size() != 0))
+								{
+									s[0].clear(), s[1].clear(), s[2].clear();
+								}
 							}
 							else
 								spiel_dran += 1;
@@ -1885,7 +1958,7 @@ int skat_game
 						if (spiel_status > 0)
 						{
 							std::cout << "><><>< " << _("player") << " \"" << 
-								pkr.key[spiel_allein].name << "\" (";
+								pkr.keys[spiel_allein].name << "\" (";
 							if ((reiz_status - 412) == 0)
 								std::cout << "VH";
 							if ((reiz_status - 412) == 1)
@@ -1900,11 +1973,11 @@ int skat_game
 								TMCG_OpenStack<VTMF_Card> os_sp2;
 								os_sp2.push(os_sp[i].first, os_sp[i].second);
 								skat_blatt(99, os_sp2);
-								std::cout << "(" << pkr.key[spiel_who[i]].name << ") ";
+								std::cout << "(" << pkr.keys[spiel_who[i]].name << ") ";
 							}
 							if (os_sp.size() < 3)
 							{
-								std::cout << " [" << pkr.key[spiel_who[spiel_dran]].name << 
+								std::cout << " [" << pkr.keys[spiel_who[spiel_dran]].name << 
 									" " << _("has to play") << "]" << std::endl;
 							}
 						}
@@ -1971,7 +2044,7 @@ int skat_game
 						if (pass_ok)
 						{
 							std::cout << "><><>< " << _("player") << " \"" << 
-								pkr.key[pkr_self].name << "\" " << _("passes at") << " " << 
+								pkr.keys[pkr_self].name << "\" " << _("passes at") << " " << 
 								reiz_wert[reiz_counter] << std::endl;
 							*out_pipe << "PRIVMSG #openSkat_" << nr << " :PASSE " << 
 								hex_game_digest << std::endl << std::flush;
@@ -2050,7 +2123,7 @@ int skat_game
 						if (bid_ok)
 						{
 							std::cout << "><><>< " << _("player") << " \"" << 
-								pkr.key[pkr_self].name << "\" " << _("bids") << " " << 
+								pkr.keys[pkr_self].name << "\" " << _("bids") << " " << 
 								reiz_wert[reiz_counter] << std::endl;
 							*out_pipe << "PRIVMSG #openSkat_" << nr << " :REIZE " << 
 								reiz_wert[reiz_counter] << " " << hex_game_digest << 
@@ -2074,7 +2147,7 @@ int skat_game
 							if ((reiz_status - 112) == ((pkr_self + p) % 3))
 							{
 								std::cout << "><><>< " << _("player") << " \"" << 
-									pkr.key[pkr_self].name << "\" " <<
+									pkr.keys[pkr_self].name << "\" " <<
 									_("doesn't take the Skat") << std::endl;
 								*out_pipe << "PRIVMSG #openSkat_" << nr << " :HAND " << 
 									hex_game_digest << std::endl << std::flush;
@@ -2097,7 +2170,7 @@ int skat_game
 							if (pkr_self == spiel_allein)
 							{
 								std::cout << "><><>< " << _("player") << " \"" << 
-									pkr.key[pkr_self].name << "\" " << 
+									pkr.keys[pkr_self].name << "\" " << 
 									_("takes the Skat") << std::endl;
 								*out_pipe << "PRIVMSG #openSkat_" << nr << " :SKAT " << 
 									hex_game_digest << std::endl << std::flush;
@@ -2163,7 +2236,7 @@ int skat_game
 												s[pkr_self].clear();
 												s[pkr_self].push(os);
 												std::cout << "><><>< " << _("player") << " \"" << 
-													pkr.key[pkr_self].name << 
+													pkr.keys[pkr_self].name << 
 													"\" " << _("pushes") << ": " << 
 													skat_type2string(tt1) << skat_type2string(tt2) << 
 													std::endl;
@@ -2265,7 +2338,7 @@ int skat_game
 												}
 											}
 											std::cout << "><><>< " << _("player") << " \"" << 
-												pkr.key[pkr_self].name << "\" " << _("announces") << 
+												pkr.keys[pkr_self].name << "\" " << _("announces") << 
 												": " << skat_spiel2string(spiel_status) << std::endl;
 											if (pctl)
 												*out_ctl << nicks[pkr_self] << " SAGEAN " <<
@@ -2321,7 +2394,7 @@ int skat_game
 											continue;
 										}
 										std::cout << "><><>< " << _("player") << " \"" << 
-											pkr.key[pkr_self].name << "\" " << _("plays the card") <<
+											pkr.keys[pkr_self].name << "\" " << _("plays the card") <<
 											": " << skat_type2string(tt) << std::endl;
 										*out_pipe << "PRIVMSG #openSkat_" << nr << " :LEGE " <<
 											skat_type2string(tt) << " " << hex_game_digest <<
@@ -2346,7 +2419,7 @@ int skat_game
 											int bk = skat_bstich(os_sp, spiel_status);
 											assert(bk != -1);
 											std::cout << "><><>< " << _("player") << " \"" << 
-												pkr.key[spiel_who[bk]].name << 
+												pkr.keys[spiel_who[bk]].name << 
 												"\" " << _("gets the trick") << ": ";
 											for (size_t i = 0; i < os_sp.size(); i++)
 												std::cout << skat_type2string(os_sp[i].first);
@@ -2369,6 +2442,12 @@ int skat_game
 											spiel_who[1] = (spiel_who[0] + 1) % 3;
 											spiel_who[2] = (spiel_who[0] + 2) % 3;
 											spiel_dran = 0;
+											// Null-Spiele ggf. sofort abbrechen
+											if ((skat_spiel2gwert(spiel_status) == 23)
+												&& (os_pkt[spiel_allein].size() != 0))
+											{
+												s[0].clear(), s[1].clear(), s[2].clear();
+											}
 										}
 										else
 											spiel_dran += 1;
@@ -2426,7 +2505,7 @@ int skat_game
 						pkt_gegner += pkt_wert[os_pkt[i][j].first];
 				pkt_allein = 120 - pkt_gegner;
 				std::cout << "><><>< " << _("card points") << " " <<
-					_("playing party") << " (" << pkr.key[spiel_allein].name << "): " << 
+					_("playing party") << " (" << pkr.keys[spiel_allein].name << "): " << 
 					pkt_allein << ", " << _("card points") << " " << 
 					_("opponent party") << ": " << pkt_gegner << std::endl;
 				
@@ -2437,11 +2516,15 @@ int skat_game
 				
 				// nachtraegliche Regelkontrolle
 				bool rules_ok[3];
-				assert(os_st.size() == 10);
+				assert((os_st.size() == 10) ||
+					((skat_spiel2gwert(spiel_status) == 23)
+						&& (os_pkt[spiel_allein].size() != 0)));
 				for (size_t i = 0; i < 3; i++)
 				{
 					TMCG_OpenStack<VTMF_Card> gps;
-					assert(os_rc[i].size() == 10);
+					assert((os_rc[i].size() == 10) ||
+						((skat_spiel2gwert(spiel_status) == 23)
+							&& (os_pkt[spiel_allein].size() != 0)));
 					gps.push(os_rc[i]);
 					rules_ok[i] = true;
 					for (size_t j = 0; j < os_st.size(); j++)
@@ -2562,7 +2645,7 @@ int skat_game
 				{
 					if (!rules_ok[i])
 					{
-						std::cout << "><>< " << _("player") << " \"" << pkr.key[i].name << 
+						std::cout << "><>< " << _("player") << " \"" << pkr.keys[i].name << 
 							"\" " << _("has violated the rules.") << std::endl;
 						return 20;
 					}
@@ -2618,7 +2701,7 @@ int skat_game
 			
 			std::cout << "><><>< " << _("game result") << " <><><> ";
 			for (size_t i = 0; i < 3; i++)
-				std::cout << pkr.key[i].name << ": " << pkt_sum[i] << " ";
+				std::cout << pkr.keys[i].name << ": " << pkt_sum[i] << " ";
 			std::cout << std::endl;
 		}
 		
@@ -2633,18 +2716,18 @@ int skat_game
 			*left << sig << std::endl << std::flush;
 			*right << sig << std::endl << std::flush;
 			left->getline(stmp, sizeof(stmp));
-			if (!pkr.key[1].verify(sig_data, stmp))
+			if (!pkr.keys[1].verify(sig_data, stmp))
 			{
 				std::cout << "><>< " << _("Signature of") << " " << _("player") <<
-					" \"" << pkr.key[1].name << "\" " << _("is invalid") << std::endl;
+					" \"" << pkr.keys[1].name << "\" " << _("is invalid") << std::endl;
 				return 30;
 			}
 			sig_protokoll << stmp << "#";
 			right->getline(stmp, sizeof(stmp));
-			if (!pkr.key[2].verify(sig_data, stmp))
+			if (!pkr.keys[2].verify(sig_data, stmp))
 			{
 				std::cout << "><>< " << _("Signature of") << " " << _("player") <<
-					" \"" << pkr.key[2].name << "\" " << _("is invalid") << std::endl;
+					" \"" << pkr.keys[2].name << "\" " << _("is invalid") << std::endl;
 				return 30;
 			}
 			sig_protokoll << stmp << "#";
@@ -2652,10 +2735,10 @@ int skat_game
 		else if (pkr_self == 1)
 		{
 			right->getline(stmp, sizeof(stmp));
-			if (!pkr.key[0].verify(sig_data, stmp))
+			if (!pkr.keys[0].verify(sig_data, stmp))
 			{
 				std::cout << "><>< " << _("Signature of") << " " << _("player") <<
-					" \"" << pkr.key[0].name << "\" " << _("is invalid") << std::endl;
+					" \"" << pkr.keys[0].name << "\" " << _("is invalid") << std::endl;
 				return 30;
 			}
 			sig_protokoll << stmp << "#";
@@ -2664,10 +2747,10 @@ int skat_game
 			*left << sig << std::endl << std::flush;
 			*right << sig << std::endl << std::flush;
 			left->getline(stmp, sizeof(stmp));
-			if (!pkr.key[2].verify(sig_data, stmp))
+			if (!pkr.keys[2].verify(sig_data, stmp))
 			{
 				std::cout << "><>< " << _("Signature of") << " " << _("player") <<
-					" \"" << pkr.key[2].name << "\" " << _("is invalid") << std::endl;
+					" \"" << pkr.keys[2].name << "\" " << _("is invalid") << std::endl;
 				return 30;
 			}
 			sig_protokoll << stmp << "#";
@@ -2675,18 +2758,18 @@ int skat_game
 		else if (pkr_self == 2)
 		{
 			left->getline(stmp, sizeof(stmp));
-			if (!pkr.key[0].verify(sig_data, stmp))
+			if (!pkr.keys[0].verify(sig_data, stmp))
 			{
 				std::cout << "><>< " << _("Signature of") << " " << _("player") <<
-					" \"" << pkr.key[0].name << "\" " << _("is invalid") << std::endl;
+					" \"" << pkr.keys[0].name << "\" " << _("is invalid") << std::endl;
 				return 30;
 			}
 			sig_protokoll << stmp << "#";
 			right->getline(stmp, sizeof(stmp));
-			if (!pkr.key[1].verify(sig_data, stmp))
+			if (!pkr.keys[1].verify(sig_data, stmp))
 			{
 				std::cout << "><>< " << _("Signature of") << " " << _("player") <<
-					" \"" << pkr.key[1].name << "\" " << _("is invalid") << std::endl;
+					" \"" << pkr.keys[1].name << "\" " << _("is invalid") << std::endl;
 				return 30;
 			}
 			sig_protokoll << stmp << "#";
@@ -2715,7 +2798,7 @@ int skat_game
 		delete [] hex_rnk_digest;
 		delete npipe;
 	}
-	delete vtmf;
+	delete vsshe,	delete vtmf;
 	if (pctl)
 		*out_ctl << nicks[pkr_self] << " DONE" << std::endl << std::flush;
 	delete out_pipe;
